unit KWARD_Host;

interface

uses
  Winapi.Windows, Winapi.Messages, System.SysUtils, System.Variants, System.Classes, Vcl.Graphics,
  Vcl.Controls, Vcl.Forms, Vcl.Dialogs, System.ImageList, Vcl.ImgList, Vcl.ExtCtrls, ShellNotifyIcon,
  Vcl.Menus;

type
  THostForm = class(TForm)
    TrayImages: TImageList;
    TrayIcon: TShellNotifyIcon;
    TrayMenu: TPopupMenu;
    WhiteIconMnu: TMenuItem;
    UnifyingMnu: TMenuItem;
    TrayN1Mnu: TMenuItem;
    LangbarMnu: TMenuItem;
    ConfigGeneralMnu: TMenuItem;
    ConfigRegeisterMnu: TMenuItem;
    TrayN2Mnu: TMenuItem;
    ExitMnu: TMenuItem;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure TrayIconShowAttempted(Sender: TObject);
    procedure TrayMenuPopup(Sender: TObject);
    procedure WhiteIconMnuClick(Sender: TObject);
    procedure UnifyingMnuClick(Sender: TObject);
    procedure LangbarMnuClick(Sender: TObject);
    procedure ConfigMnusClick(Sender: TObject);
    procedure ExitMnuClick(Sender: TObject);
  private
    { Private 錾 }
    FOpenStatus: Boolean;
    FConversion: Integer;
    Fx64Handle: HWND;
    procedure ReadOption;
    function Monitored: Boolean;
    procedure MonitorStart;
    procedure MonitorStop;
    function GetImeStatus(Open: Boolean; Conversion: Integer): Integer;
    procedure UpdateIconIndex;
  protected
    { Protected 錾 }
    procedure WndProc(var Message: TMessage); override;
  public
    { Public 錾 }
  end;

var
  HostForm: THostForm;

implementation

{$R *.dfm}

uses
  RegData, SysInfo, Winapi.ShellAPI, WndInfo, Winapi.Imm;

type
  //KanaWard.dll ֐
  TInstallHook = function: Boolean; stdcall;
  TUninstallHook = procedure; stdcall;

const
  ImeNotifyID = '{5C8D3F6E-80CD-4F50-BB5A-578C21ED49C3}';

  //LanguageBar ֘A̒ǉB
  SPI_GETSYSTEMLANGUAGEBAR = $1050;
  SPI_SETSYSTEMLANGUAGEBAR = $1051;
  SPIF_INICHANGE = SPIF_UPDATEINIFILE or SPIF_SENDCHANGE;

resourcestring
  SNotFound = '"%s" ܂';

var
  LibHandle: THandle;
  InstallHook: TInstallHook;
  UninstallHook: TUninstallHook;

  { IME ̏ԕύX󂯂郆[U[bZ[WB
    WPARAM: ImmGetOpenStatus ̕ԂlƓ
    LPARAM: ImmGetConversionStatus  lpfdwConversion Ɠ }
  ImeNotifyMsg: UINT;

{ THostForm }

procedure THostForm.ReadOption;
var
  Reg: TRegData;
begin
  Reg := TRegData.Create;
  try
    if Reg.OpenKey('Software\ShaggyDog\KanaWard\Option', True) then
    begin
      WhiteIconMnu.Checked := Reg.ReadBoolDef('WhiteIcon', True);
      UnifyingMnu.Checked := Reg.ReadBoolDef('Unifying', False);
      Reg.CloseKey;
    end;
  finally
    Reg.Free;
  end;
end;

function THostForm.Monitored: Boolean;
begin
  if IsWin64 then
    Result := Fx64Handle > 0
  else
    Result := LibHandle > 0;
end;

procedure THostForm.MonitorStart;
var
  FileName: string;
  Info: TShellExecuteInfo;
begin
  if Monitored then Exit;

  //tbNJn
  LibHandle := LoadLibrary(PChar(ChangeFileExt(ParamStr(0), '.dll')));
  if LibHandle <> 0 then
  begin
    @InstallHook := GetProcAddress(LibHandle, 'InstallHook');
    if not Assigned(@InstallHook) then
      raise Exception.Create(SysErrorMessage(GetLastError));
    @UninstallHook := GetProcAddress(LibHandle, 'UninstallHook');
    if not Assigned(@UninstallHook) then
      raise Exception.Create(SysErrorMessage(GetLastError));

    if not InstallHook then
      raise Exception.Create(SysErrorMessage(GetLastError));
  end
  else
    raise Exception.Create(SysErrorMessage(GetLastError));

  //x64 N
  if IsWin64 then
  begin
    FileName := ChangeFileExt(ParamStr(0), '.x64.exe');
    if FileExists(FileName) then
    begin
      FillChar(Info, SizeOf(TShellExecuteInfo), 0);
      with Info do begin
        cbSize := SizeOf(Info);
        fMask := SEE_MASK_NOCLOSEPROCESS;
        lpVerb := 'open';
        lpFile := PChar(FileName);
        lpParameters := nil;
        lpDirectory := nil;
        nShow := SW_SHOW;
      end;
      if ShellExecuteEx(@Info) then
      begin
        WaitForInputIdle(Info.hProcess, INFINITE);  //Ô߁B
        Fx64Handle := FindWindowIn(GetProcessId(Info.hProcess), 'TEscortForm', 'Kana Ward (x64)');  //WindowsXP SP1 ȍ~B
        CloseHandle(Info.hProcess);
      end;
    end
    else
      raise EFileNotFoundException.Create(Format(SNotFound, [FileName]));
  end;
end;

procedure THostForm.MonitorStop;
begin
  if Fx64Handle <> 0 then
  begin
    PostMessage(Fx64Handle, WM_CLOSE, 0, 0);
    Fx64Handle := 0;
  end;

  UnInstallHook;
  FreeLibrary(LibHandle);
  LibHandle := 0;
end;

procedure THostForm.FormCreate(Sender: TObject);
var
  Reg: TRegData;
begin
  ReadOption;
  FConversion := 0;
  FOpenStatus := False;
  Fx64Handle := 0;

  Application.ShowMainForm := False;
  TrayIcon.Show;

  ImeNotifyMsg := RegisterWindowMessage(ImeNotifyID);
  if ImeNotifyMsg <> 0 then
  begin
    Reg := TRegData.Create;
    try
      if Reg.OpenKey('Software\ShaggyDog\KanaWard\Variable', True) then
      begin
        Reg.WriteInteger('HostWnd', Self.Handle);
        Reg.WriteInteger('ImeNotifyMsg', ImeNotifyMsg);
        Reg.WriteBool('OpenStatus', FOpenStatus);
        Reg.WriteInteger('Conversion', FConversion);
        Reg.CloseKey;
      end;
    finally
      Reg.Free;
    end;
    if TOSVersion.Check(6) then
      ChangeWindowMessageFilter(ImeNotifyMsg, MSGFLT_ADD);
    MonitorStart;
  end
  else
    raise Exception.Create(SysErrorMessage(GetLastError));
end;

//IME ̏ԂACRCfbNX𓾂B
//TSF łԂ̓ǂݏo Imm n API ƃbZ[WgB
function THostForm.GetImeStatus(Open: Boolean; Conversion: Integer): Integer;
var
  Mode: Integer;
begin
  if Open then
  begin
    Mode := Conversion and IME_CMODE_ALPHANUMERIC;
    Mode := Mode or (Conversion and IME_CMODE_NATIVE);
    Mode := Mode or (Conversion and IME_CMODE_KATAKANA);
    Mode := Mode or (Conversion and IME_CMODE_FULLSHAPE);
    case Mode of
      IME_CMODE_NATIVE or IME_CMODE_FULLSHAPE:
        Result := 1;  //SpЂ炪
      IME_CMODE_NATIVE or IME_CMODE_FULLSHAPE or IME_CMODE_KATAKANA:
        Result := 2;  //SpJ^Ji
      IME_CMODE_NATIVE or IME_CMODE_KATAKANA:
        Result := 3;  //pJ^Ji
      IME_CMODE_FULLSHAPE:
        if (GetKeyState(VK_CAPITAL) and 1) > 0 then
          Result := 4 //Sppi啶j
        else
          Result := 6;  //Sppij
      IME_CMODE_ALPHANUMERIC:
        if (GetKeyState(VK_CAPITAL) and 1) > 0 then
          Result := 5   //ppi啶j
        else
          Result := 7;  //ppij
    else
      Result := 0;
    end;
  end
  else
  begin
    if (GetKeyState(VK_CAPITAL) and 1) > 0 then
      Result := 5   //ppi啶j
    else
      Result := 7;  //ppij
    if (GetKeyState(VK_KANA) and 1) > 0 then
      Result := 3;  //pJ^Ji
  end;
end;

procedure THostForm.WndProc(var Message: TMessage);
var
  Reg: TRegData;
begin
  inherited WndProc(Message);

  if Message.Msg = ImeNotifyMsg then
  begin
    if not TrayIcon.Visible then
      TrayIcon.Show;

    FOpenStatus := Message.WParam <> 0;
    FConversion := Message.LParam;
    TrayIcon.IconIndex := GetImeStatus(FOpenStatus, FConversion);

    if UnifyingMnu.Checked then
    begin
      Reg := TRegData.Create;
      try
        if Reg.OpenKey('Software\ShaggyDog\KanaWard\Variable', True) then
        begin
          Reg.WriteBool('OpenStatus', FOpenStatus);
          Reg.WriteInteger('Conversion', FConversion);
          Reg.CloseKey;
        end;
      finally
        Reg.Free;
      end;
    end;
  end;
end;

procedure THostForm.FormDestroy(Sender: TObject);
begin
  MonitorStop;
end;

procedure THostForm.UpdateIconIndex;
begin
  if WhiteIconMnu.Checked then
    TrayImages.ImageType := itImage
  else
    TrayImages.ImageType := itMask;
  TrayIcon.IconIndex := GetImeStatus(FOpenStatus, FConversion)
end;

procedure THostForm.TrayIconShowAttempted(Sender: TObject);
var
  Imc: HIMC;
  Conversion,
  Sentence: DWORD;
begin
  Imc := ImmGetContext(Handle);
  try
    if Imc <> 0 then
    begin
      FOpenStatus := ImmGetOpenStatus(Imc);
      ImmGetConversionStatus(Imc, Conversion, Sentence);
      FConversion := Conversion;

      UpdateIconIndex;
    end;
  finally
    ImmReleaseContext(Handle, Imc);
  end;
end;

procedure THostForm.TrayMenuPopup(Sender: TObject);
var
  Shown: BOOL;
begin
  LangbarMnu.Visible := SystemParametersInfo(SPI_GETSYSTEMLANGUAGEBAR, 0, @Shown, 0);
  if LangbarMnu.Visible then
    LangbarMnu.Checked := Shown;
end;

procedure THostForm.WhiteIconMnuClick(Sender: TObject);
var
  Reg: TRegData;
begin
  TrayIcon.IconIndex := -1;
  UpdateIconIndex;
  Reg := TRegData.Create;
  try
    if Reg.OpenKey('Software\ShaggyDog\KanaWard\Option', True) then
    begin
      Reg.WriteBool('WhiteIcon', WhiteIconMnu.Checked);
      Reg.CloseKey;
    end;
  finally
    Reg.Free;
  end;
end;

procedure THostForm.UnifyingMnuClick(Sender: TObject);
var
  Reg: TRegData;
begin
  MonitorStop;
  try
    SendNotifyMessage(HWND_BROADCAST, WM_NULL, 0, 0);

    Reg := TRegData.Create;
    try
      if Reg.OpenKey('Software\ShaggyDog\KanaWard\Option', True) then
        Reg.WriteBool('Unifying', UnifyingMnu.Checked);
      Reg.CloseKey;
    finally
      Reg.Free;
    end;
  finally
    MonitorStart;
  end;
end;

procedure THostForm.LangbarMnuClick(Sender: TObject);
begin
  //Win8 ȍ~BȑOł̐@͌ȂB
  SystemParametersInfo(SPI_SETSYSTEMLANGUAGEBAR, 0, Pointer(LangbarMnu.Checked), SPIF_INICHANGE);
end;

procedure THostForm.ConfigMnusClick(Sender: TObject);
begin
  // Imm n APIBWinapi.MsCTF gȂB
  ImmConfigureIME(GetKeyboardLayout(0), Handle, TMenuItem(Sender).Tag, nil);
end;

procedure THostForm.ExitMnuClick(Sender: TObject);
begin
  Close;
end;

end.
